#include <iostream>

#include "risc.h"
#include <list>

typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;

void get_all_instances_of(std::list<risc::InstanceTree> &instances,
         risc::InstanceTree root, risc::Definition::Type type, bool first);

risc::sg::Time
get_waiting_time(risc::sg::SegmentGraph &segment_graph, VertexIterator vertex_iter);

class EarlistExecutionTable {

public:

  EarlistExecutionTable(risc::sg::SegmentGraph &segment_graph):
    segment_graph_(segment_graph)
  {
    size_ = boost::num_vertices(segment_graph_.graph_);
    table_ = new risc::sg::Time[size_];

    std::pair<VertexIterator, VertexIterator> vp;
    for(vp = boost::vertices(segment_graph_.graph_);
        vp.first != vp.second; ++vp.first) {

      table_index_segment_lookup_.push_back(vp.first);
    }

    determine_table();
  }

  ~EarlistExecutionTable()
  {
    delete[] table_;
  }

  void determine_table();

  void print_table();

  risc::sg::Time& get_min_time(VertexIterator vertex_iter);

  /**
   * \brief    This vector is a lookup between the conflict table index and the
   *           associated segment index
   */
  std::vector<VertexIterator> table_index_segment_lookup_;

  int size_;
  risc::sg::Time *table_;
  risc::sg::SegmentGraph &segment_graph_;
};

void
EarlistExecutionTable::print_table()
{
  for(int i=0; i<size_; i++) {

    std::cout << segment_graph_.graph_[*table_index_segment_lookup_.at(i)].id_
              << ", "
              << table_[i]
              << std::endl;
  }
}

risc::sg::Time&
EarlistExecutionTable::get_min_time(VertexIterator vertex_iter)
{
  for(int i = 0; i < table_index_segment_lookup_.size(); i++) {

    if(table_index_segment_lookup_.at(i) == vertex_iter) {

      return table_[i];
    }
  }
}

void
EarlistExecutionTable::determine_table()
{
  for(int i=0; i<size_; i++) {

    typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;
    std::pair<VertexIterator, VertexIterator> vp;

    VertexIterator temp;

    for(vp = boost::vertices(segment_graph_.graph_);
        vp.first != vp.second;
        ++vp.first) {

      if(segment_graph_.graph_[*(vp.first)].color_for_communication_graph_
          == risc::sg::Segment::Touched) {

        continue;
      }

      risc::sg::Graph::in_edge_iterator in_edge_begin, in_edge_end, in_edge_iter;
      boost::tie(in_edge_begin, in_edge_end)
        = boost::in_edges(*(vp.first), segment_graph_.graph_);

      if(in_edge_begin == in_edge_end) {
        temp = (vp.first);

        for(int i = 0; i < table_index_segment_lookup_.size(); i++) {
          if(table_index_segment_lookup_.at(i) == vp.first) {

            risc::sg::Time time;
            table_[i] = time;
          }
        }

        break;
      }

      bool all_predecessor_touched = true;
      for(in_edge_iter  = in_edge_begin;
          in_edge_iter != in_edge_end;
          in_edge_iter++) {

        if(segment_graph_.graph_[
            boost::source(*in_edge_iter, segment_graph_.graph_)].
              color_for_communication_graph_
            == risc::sg::Segment::Untouched) {

          all_predecessor_touched = false;
        }
      }

      if(all_predecessor_touched) {

        temp = (vp.first);
        break;
      }

      temp = (vp.first);
    }

    segment_graph_.graph_[*temp].color_for_communication_graph_
      = risc::sg::Segment::Touched;

    risc::sg::Graph::in_edge_iterator in_edge_begin, in_edge_end, in_edge_iter;
    boost::tie(in_edge_begin, in_edge_end)
      = boost::in_edges(*temp, segment_graph_.graph_);



    risc::sg::Time min_time;
    bool first_iteration = true;
    for(in_edge_iter  = in_edge_begin;
        in_edge_iter != in_edge_end;
        in_edge_iter++) {

      risc::sg::Time waiting_time_for_segment =
        get_waiting_time(segment_graph_, (vp.first));

      if(first_iteration) {

        first_iteration = false;
        min_time = get_min_time(
                       std::find(vertices(segment_graph_.graph_).first,
                                vertices(segment_graph_.graph_).second,
                                boost::source(*in_edge_iter,
                                           segment_graph_.graph_)));

        min_time = min_time + waiting_time_for_segment;

      } else {

        risc::sg::Time temp = get_min_time(
                        std::find(vertices(segment_graph_.graph_).first,
                                 vertices(segment_graph_.graph_).second,
                                 boost::source(*in_edge_iter,
                                            segment_graph_.graph_)));

        if(min_time + waiting_time_for_segment < min_time) {
          min_time = temp + waiting_time_for_segment;
        } else {
          //std::cout << "ouch" << std::endl;
        }
      }
    }

    get_min_time(vp.first) = min_time;
  }
}

risc::sg::Time
get_waiting_time(risc::sg::SegmentGraph &segment_graph, VertexIterator vertex_iter)
{
  segment_graph.graph_[*vertex_iter].boundary_node_;

  const SgStatement *stmt = isSgStatement(segment_graph.graph_[*vertex_iter].boundary_node_);
  const SgExprStatement *expr_stmt = isSgExprStatement(stmt);

  if(expr_stmt) {

    const SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    if(func_call) {

      if(SageInterface::get_name(
            func_call->getAssociatedFunctionDeclaration()) == "wait") {

        if(func_call->get_args()->get_expressions().size() == 1) {

          if(func_call->get_args()-> get_expressions().at(0)
              ->unparseToString() == "SC_ZERO_TIME") {

            return risc::sg::Time(0,1);

          } else {

            SgVarRefExp* waiting_event
              = isSgVarRefExp(isSgArrowExp(func_call->get_args()->
                    get_expressions().at(0))->get_rhs_operand());

            if(risc::tools::is_sc_event(waiting_event)) {

              return risc::sg::Time(0,1);
            } else {
              std::cout << "ERROR1" << std::endl;
            }
          }
        }

        if(func_call->get_args()->get_expressions().size() == 0) {

            return risc::sg::Time(0,0);
        }
      }
    }
  }
  return risc::sg::Time(0,0);
}



int main(int argc, char* argv[])
{
  std::cout << std::endl
            << "start processing file: " __FILE__
            << " ..."
            << std::endl;

  risc::Design d(argc,argv);

  std::cout << "Root: ";
  d.get_root().print_instance_tree();


  //////////////////////////////////////////////////////////
  std::list<risc::InstanceTree> module_instances;
  get_all_instances_of(module_instances,
      d.get_root(),
      risc::Definition::ModuleInstanceType,
      true);

  std::cout << "Module Instances:" << std::endl;
  for(std::list<risc::InstanceTree>::iterator
      iter  = module_instances.begin();
      iter != module_instances.end();
      iter++) {
    std::cout << "  ";
    (*iter).print_instance_tree();
  }


  //////////////////////////////////////////////////////////
  std::list<risc::InstanceTree> event_instances;
  get_all_instances_of(event_instances,
      d.get_root(),
      risc::Definition::EventType,
      true);

  std::cout << "Event Instances:" << std::endl;
  for(std::list<risc::InstanceTree>::iterator
      iter  = event_instances.begin();
      iter != event_instances.end();
      iter++) {
    std::cout << "  ";
    (*iter).print_instance_tree();
  }


  //////////////////////////////////////////////////////////
  risc::sg::SegmentGraph::FunctionCallBoundarySet
    function_call_boundary(d.systemc_definitions_->sc_wait_funcs_);
  risc::sg::SegmentGraph sg(function_call_boundary);

  for(std::list<risc::InstanceTree>::iterator
      iter  = module_instances.begin();
      iter != module_instances.end();
      iter++) {

    risc::ModuleInstance *module_instance
      = dynamic_cast<risc::ModuleInstance*>((*iter).get_object());

    for(risc::ThreadVectorIter
        iter_thread  = module_instance->definition_->threads_.begin();
        iter_thread != module_instance->definition_->threads_.end();
        iter_thread++) {

      if((*iter_thread)->get_name() == "thread1" ) {

        AstDOTGeneration dotgen;
        dotgen.generate((*iter_thread)->get_ast_node(), (*iter_thread)->get_name());
      }



      std::cout << (*iter_thread)->get_name() << std::endl;

      sg.add_function(*iter_thread, true);
    }
  }

  risc::sg::EventConflictTable event_conflicts(sg);

  sg.clean_graph();

  EarlistExecutionTable execution_table(sg);

  for(int row=0; row < event_conflicts.size_; row++) {

    for(int column=0; column < event_conflicts.size_; column++) {

      if(event_conflicts.table_[row][column]) {

        std::cout << "Row ID: "
                  << sg.graph_[*(execution_table.table_index_segment_lookup_[row])].id_
                  << std::endl;

        std::cout << "Column ID: "
                  << sg.graph_[*(execution_table.table_index_segment_lookup_[column])].id_
                  << std::endl;

        std::cout << std::endl;

        if(
        execution_table.get_min_time(execution_table.table_index_segment_lookup_[row]).
          one_delta_cycle_ahead(
            execution_table.get_min_time(execution_table.table_index_segment_lookup_[column]))) {

          std::cout << "adding: "
                    << sg.graph_[*event_conflicts.table_index_segment_lookup_.at(row)].id_ 
                    << " -> "
                    << sg.graph_[*event_conflicts.table_index_segment_lookup_.at(column)].id_ 
                    << std::endl;

          boost::add_edge(
              *event_conflicts.table_index_segment_lookup_.at(row),
              *event_conflicts.table_index_segment_lookup_.at(column),
              sg.graph_);
        }
      }
    }
  }

  sg.print_graph("communication_graph.dot");
  event_conflicts.print_event_conflict_table("event_conflict_table.html");



  std::cout << std::endl;
  std::cout << "Print table" << std::endl;
  execution_table.print_table();

  std::cout << std::endl;


  std::cout << "end processing file: " __FILE__
            << " ..."
            << std::endl
            << std::endl;

  return 0;
}

// Assumption: only channels in channels and modules in modules

void get_all_instances_of(std::list<risc::InstanceTree> &instances,
         risc::InstanceTree root, risc::Definition::Type type, bool first)
{
  risc::ModuleInstance *module_instance
    = dynamic_cast<risc::ModuleInstance*>(root.get_object());

  if(module_instance) {

    for(risc::ModuleInstanceVectorIter
        iter  = module_instance->definition_->sub_modules_.begin();
        iter != module_instance->definition_->sub_modules_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->definition_;
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    for(risc::HierarchicalChannelInstanceVectorIter
        iter  = module_instance->definition_->sub_hier_channel_.begin();
        iter != module_instance->definition_->sub_hier_channel_.end();
        iter++) {

      risc::InstanceTree::Element sub_path;

      sub_path.first = module_instance->definition_;
      sub_path.second = *iter;

      root.push_back(sub_path);
      get_all_instances_of(instances, root, type, false);
      root.pop_back();
    }

    if(type == risc::Definition::EventType) {
      for(risc::EventVectorIter
          iter  = module_instance->definition_->events_.begin();
          iter != module_instance->definition_->events_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->definition_, *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }

    if(type == risc::Definition::ModuleInstanceType) {

      if(first) {
        instances.push_back(root);
      }

      for(risc::ModuleInstanceVectorIter
          iter  = module_instance->definition_->sub_modules_.begin();
          iter != module_instance->definition_->sub_modules_.end();
          iter++) {

        risc::InstanceTree temp = root;
        risc::InstanceTree::Element e(module_instance->definition_, *iter);
        temp.push_back(e);
        instances.push_back(temp);
      }
    }
  }






//  ModuleInstance *module_instance
//    = dynamic_cast<ModuleInstance*>(root.get_object());
//
//  if(module_instance) {
//
//    for(ModuleInstanceVectorIter
//        iter  = module_instance->definition_->sub_modules_.begin();
//        iter != module_instance->definition_->sub_modules_.end();
//        iter++) {
//
//      Element sub_path;
//
//      sub_path.first = module_instance->definition_;
//      sub_path.second = *iter;
//
//      path_.push_back(sub_path);
//      get_all_ports(occurence);
//      path_.pop_back();
//    }
//
//    for(HierarchicalChannelInstanceVectorIter
//        iter  = module_instance->definition_->sub_hier_channel_.begin();
//        iter != module_instance->definition_->sub_hier_channel_.end();
//        iter++) {
//
//      Element sub_path;
//
//      sub_path.first = module_instance->definition_;
//      sub_path.second = *iter;
//
//      path_.push_back(sub_path);
//      get_all_ports(occurence);
//      path_.pop_back();
//    }
//
//    for(EventVectorIter
//        iter  = module_instance->definition_->events_.begin();
//        iter != module_instance->definition_->events_.end();
//        iter++) {
//
//      InstanceTree temp = *this;
//      Element e(module_instance->definition_, *iter);
//      temp.path_.push_back(e);
//      occurence.push_back(temp);
//    }
//  }

//  HierarchicalChannelInstance *hier_channel_instance
//    = dynamic_cast<HierarchicalChannelInstance*>(get_object());
//
//  if(hier_channel_instance) {
//
//    for(ModuleInstanceVectorIter
//        iter  = hier_channel_instance->definition_->sub_modules_.begin();
//        iter != hier_channel_instance->definition_->sub_modules_.end();
//        iter++) {
//
//      Element sub_path;
//
//      sub_path.first = hier_channel_instance->definition_;
//      sub_path.second = *iter;
//
//      path_.push_back(sub_path);
//      get_all_ports(occurence);
//      path_.pop_back();
//    }
//
//    for(HierarchicalChannelInstanceVectorIter
//        iter  = hier_channel_instance->definition_->sub_hier_channel_.begin();
//        iter != hier_channel_instance->definition_->sub_hier_channel_.end();
//        iter++) {
//
//      Element sub_path;
//
//      sub_path.first = hier_channel_instance->definition_;
//      sub_path.second = *iter;
//
//      path_.push_back(sub_path);
//      get_all_ports(occurence);
//      path_.pop_back();
//    }
//
//    for(EventVectorIter
//        iter  = hier_channel_instance->definition_->events_.begin();
//        iter != hier_channel_instance->definition_->events_.end();
//        iter++) {
//
//      InstanceTree temp = *this;
//      Element e(hier_channel_instance->definition_, *iter);
//      temp.path_.push_back(e);
//      occurence.push_back(temp);
//    }
//  }
//
//  PrimitiveChannelInstance *prim_channel_instance
//    = dynamic_cast<PrimitiveChannelInstance*>(get_object());
//
//  if(prim_channel_instance) {
//
//    for(EventVectorIter
//        iter  = prim_channel_instance->definition_->events_.begin();
//        iter != prim_channel_instance->definition_->events_.end();
//        iter++) {
//
//      InstanceTree temp = *this;
//      Element e(prim_channel_instance->definition_, *iter);
//      temp.path_.push_back(e);
//      occurence.push_back(temp);
//    }
//  }
}


/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
