// mandelbrot.cpp: Mandelbrot set renderer example (based on mandelbrot.sc by RD)
// Author: Guantao Liu, CECS
// Date: 03/30/2015

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "systemc.h"


// Definitions ///////////////////////////////////////////////////////////


//#define DEBUG         // print debugging messages

//#define JPEG          // convert saved images to .jpg format

//#define DISPLAY       // display calculated image (using SDL)

//#define HELP          // print help messages

//#define float_t    float
//#define float_t    double
#define float_t    long double

// pretty location "Zom-B" from:
// http://en.wikipedia.org/wiki/File:Mandelbrot_sequence_new.gif
#define ZOM_B_X         -0.743643887037158704752191506114774L
#define ZOM_B_Y         0.131825904205311970493132056385139L

#ifndef WIDTH           // size of the image in pixels
#define WIDTH           640
#endif
#ifndef HEIGHT
#define HEIGHT          512
#endif
#ifndef MAX_ITER        // depth of calculation
//#define MAX_ITER      1024
//#define MAX_ITER      2048
//#define MAX_ITER      3072
#define MAX_ITER        4096
#endif
#ifndef START_LEFT      // initial position over the Mandelbrot set
//#define START_LEFT    -2.2
#define START_LEFT      (-1.6 + ZOM_B_X)
#endif
#ifndef START_RIGHT
//#define START_RIGHT   1.0
#define START_RIGHT     (1.6 + ZOM_B_X)
#endif
#ifndef START_BOTTOM
//#define START_BOTTOM  -1.0
#define START_BOTTOM    (-1.12 + ZOM_B_Y)
#endif
#ifndef START_TOP
//#define START_TOP     1.0
#define START_TOP       (1.12 + ZOM_B_Y)
#endif
#ifndef TARGET_COORDS_X // target coordinates
#define TARGET_COORDS_X ZOM_B_X
#endif
#ifndef TARGET_COORDS_Y
#define TARGET_COORDS_Y	ZOM_B_Y
#endif
#ifndef TARGET_ZOOM     // zoom factor per step towards the target
#define TARGET_ZOOM     0.7
#endif
#ifndef NUM_IMAGES
#define NUM_IMAGES      20     // number of images produced
#endif
#ifndef MAX_IMAGE_FILES
#define MAX_IMAGE_FILES	0     // number of image files kept
#endif

#define LINEAR_MOTION	// move in equidistant steps (constant speed)
//#define FASTER_MOTION	// move in logarithmic steps (slow, then faster and faster)
//#define SLOWER_MOTION	// move in logarithmic steps (fast, then slower and slower)

#ifndef PAR             // number of parallel units instantiated
//#define PAR           1
//#define PAR           2
//#define PAR           4
#define PAR           8
//#define PAR           16
//#define PAR           32
//#define PAR           64
#endif

#define TIME_BASE       (1 MICRO_SEC)   // print time in units of micro-seconds

#ifdef DISPLAY
#include "sdlsysc.h"
#endif /* DISPLAY */

//#define SYSC_TYPE_CHNL

#define USE_SC_THREAD


// Data structures ///////////////////////////////////////////////////////


typedef struct Image
{
    unsigned char R[HEIGHT][WIDTH];
    unsigned char G[HEIGHT][WIDTH];
    unsigned char B[HEIGHT][WIDTH];

    Image(void)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = 0;
                G[i][j] = 0;
                B[i][j] = 0;
            }
    }

    Image& operator=(const Image& copy)
    {
        for (int i=0; i<HEIGHT; i++)
            for (int j=0; j<WIDTH; j++)
            {
                R[i][j] = copy.R[i][j];
                G[i][j] = copy.G[i][j];
                B[i][j] = copy.B[i][j];
            }
        return *this;
    }

    friend ostream& operator<<(ostream& os, const Image& im);
} IMAGE;

ostream& operator<<(ostream& os, const Image& im)
{
    for (int i=0; i<HEIGHT; i++)
        for (int j=0; j<WIDTH; j++)
        {
            os << im.R[i][j] << ' ';
            os << im.G[i][j] << ' ';
            os << im.B[i][j] << '\n';
        }
    return os;
}

typedef struct Coordinates
{
    float_t l;	// left bound, e.g. -2.5
    float_t r;	// right bound, e.g. 1.0
    float_t b;	// bottom bound, e.g. -1.0
    float_t t;	// top bound, e.g. 1.0

    friend ostream& operator<<(ostream& os, const Coordinates& co);
} COORDS;

ostream& operator<<(ostream& os, const Coordinates& co)
{
    os << co.l << ' ' << co.r << ' ' << co.b << ' ' << co.t << '\n';
    return os;
}


// Channels //////////////////////////////////////////////////////////////


#ifdef SYSC_TYPE_CHNL // use SystemC-style FIFO channels
typedef sc_fifo_out_if<IMAGE> i_image_sender;
typedef sc_fifo_in_if<IMAGE> i_image_receiver;
typedef sc_fifo<IMAGE> c_image_channel;

typedef sc_fifo_out_if<COORDS> i_coords_sender;
typedef sc_fifo_in_if<COORDS> i_coords_receiver;
typedef sc_fifo<COORDS> c_coords_channel;
#else // use SpecC-style double handshake channels
class i_image_sender: virtual public sc_interface
{
    public:
        virtual void write(IMAGE&) = 0;
};

class i_image_receiver: virtual public sc_interface
{
    public:
        virtual void read(IMAGE&) = 0;
};

class c_image_channel: public sc_channel, public i_image_sender, public i_image_receiver
{
    public:
        c_image_channel(sc_module_name name, int size_ = 4): sc_channel(name)
        {
            size = size_;
            buf = new IMAGE[size];
            reset();
        }

        ~c_image_channel()
        {
            delete [] buf;
        }

        int num_available() const
        {
            return size - free_slots;
        }

        int num_free() const
        {
            return free_slots;
        }

        void write(IMAGE& X)
        {
            if (num_free() == 0)
                wait(data_read_event);
            buf[wi] = X;
            wi = (wi+1) % size;
            free_slots --;
            data_written_event.notify(SC_ZERO_TIME);
        }

        void read(IMAGE& Y)
        {
            if (num_available() == 0)
                wait(data_written_event);
            Y = buf[ri];
            ri = (ri+1) % size;
            free_slots ++;
            data_read_event.notify(SC_ZERO_TIME);
        }

        void reset()
        {
            free_slots = size;
            ri = 0;
            wi = 0;
        }

    private:
        int size;       // size
        IMAGE *buf;     // circular buffer
        int free_slots; // free space
        int ri;         // read index
        int wi;         // write index

        sc_event data_read_event;
        sc_event data_written_event;
};

class i_coords_sender: virtual public sc_interface
{
    public:
        virtual void write(COORDS&) = 0;
};

class i_coords_receiver: virtual public sc_interface
{
    public:
        virtual void read(COORDS&) = 0;
};

class c_coords_channel: public sc_channel, public i_coords_sender, public i_coords_receiver
{
    public:
        c_coords_channel(sc_module_name name, int size_ = 4): sc_channel(name)
        {
            size = size_;
            buf = new COORDS[size];
            reset();
        }

        ~c_coords_channel()
        {
            delete [] buf;
        }

        int num_available() const
        {
            return size - free_slots;
        }

        int num_free() const
        {
            return free_slots;
        }

        void write(COORDS& X)
        {
            if (num_free() == 0)
                wait(data_read_event);
            buf[wi] = X;
            wi = (wi+1) % size;
            free_slots --;
            data_written_event.notify(SC_ZERO_TIME);
        }

        void read(COORDS& Y)
        {
            if (num_available() == 0)
                wait(data_written_event);
            Y = buf[ri];
            ri = (ri+1) % size;
            free_slots ++;
            data_read_event.notify(SC_ZERO_TIME);
        }

        void reset()
        {
            free_slots = size;
            ri = 0;
            wi = 0;
        }

    private:
        int size;       // size
        COORDS *buf;     // circular buffer
        int free_slots; // free space
        int ri;         // read index
        int wi;         // write index

        sc_event data_read_event;
        sc_event data_written_event;
};
#endif /* SYSC_TYPE_CHNL */


// Modules /////////////////////////////////////////////////////////////


SC_MODULE(Monitor)
{
    sc_port<i_image_receiver> ImgIn;
    //in sim_time StartTime[NUM_IMAGES]
    IMAGE image;	// current received image

    int SaveImage(const char *Filename,
                unsigned char R[HEIGHT][WIDTH],
                unsigned char G[HEIGHT][WIDTH],
                unsigned char B[HEIGHT][WIDTH])
    {
#ifdef JPEG
    char    SysCmd[256];
#endif /* JPEG */
    FILE    *File;
    int	y, x;

    assert(Filename);
    assert(strlen(Filename) < 80);
#ifdef DEBUG
    printf("DEBUG: Saving file %s...\n", Filename);
#endif
    File = fopen(Filename, "w");
    if (!File)
    {
        printf("\nCannot open file \"%s\" for writing!\n", Filename);
        return 1;
    }
    fprintf(File, "P6\n");
    fprintf(File, "%d %d\n", WIDTH, HEIGHT);
    fprintf(File, "255\n");
    for(y=0; y<HEIGHT; y++)
    {
        for(x=0; x<WIDTH; x++)
        {
            fputc(R[y][x], File);
            fputc(G[y][x], File);
            fputc(B[y][x], File);
        }
    }
    if (ferror(File))
    {
        printf("\nFile error while writing to file!\n");
        fclose(File);
        return 2;
    }
    fclose(File);
#ifdef DEBUG
    printf("DEBUG: File %s was saved successfully.\n", Filename);
#endif
#ifdef JPEG
#ifdef DEBUG
    printf("DEBUG: Converting file %s to %s.jpg...\n", Filename, Filename);
#endif
    sprintf(SysCmd, "pnmtojpeg %s >%s.jpg", Filename, Filename);
    if (system(SysCmd) != 0)
    {
        printf("\nError while converting PPM to JPG:\nCommand \"%s\" failed!\n", SysCmd);
        return 10;
    }
#endif /* JPEG */
    return 0;
    } /* end of SaveImage */

    void main(void)
    {
#if (MAX_IMAGE_FILES > 0)
        char outfilename[128];
        int n;
#endif
        //sim_time t, t2;
        //sim_time_string buf, buf2;
        int i;

        for(i=0; i<NUM_IMAGES; i++)
        {
            ImgIn->read(image);
            //t = now();
#ifdef HELP
            printf("Monitor received image %d.\n", i);
#endif
            //t2 = t - StartTime[i];
            //if (t2 > 0)
            //{   printf("%8s: Image processing took %8s micro seconds.\n",
            //    time2str(buf, t/TIME_BASE), time2str(buf2, t2/TIME_BASE));
            //}
#if (MAX_IMAGE_FILES > 0)
            n = i % MAX_IMAGE_FILES;
            sprintf(outfilename, "mandelbrot_%03d.ppm", n);
            if (SaveImage(outfilename, image.R, image.G, image.B))
            {
                fprintf(stderr, "Error writing the image \"%s\".\n", outfilename);
                exit(10);
            }
#endif
#ifdef DISPLAY
            if (_SDL_SysC_DisplayClosed())
            {
                printf("Display closed, monitor quits.\n");
                exit(0);
            }
#endif /* DISPLAY */
        }
#ifdef HELP
        printf("Monitor quits.\n");
#endif
        exit(0);
    }

    SC_CTOR(Monitor)
    {
        SC_THREAD(main);
#ifdef SYSC_TYPE_CHNL
        set_stack_size(128*1024*1024);
#endif /* SYSC_TYPE_CHNL */
    }
};


SC_MODULE(Stimulus)
{
    sc_port<i_coords_sender> CoordsOut;
    //out sim_time StartTime[NUM_IMAGES]
    COORDS coords;

    void main(void)
    {
        //sim_time t;
        //sim_time_string buf;
#ifdef DISPLAY
        char title[] = "Mandelbrot XXXX";
#endif /* DISPLAY */
        int i = 0;
        float_t center_x, center_y;
#ifdef LINEAR_MOTION
        float_t offset_x, offset_y;
#endif
#ifdef FASTER_MOTION
        float_t start_x, start_y;
        float_t dist_x, dist_y;
        float_t reverse_x, reverse_y;
        float_t factor_x, factor_y;
        float_t product_x, product_y;
#endif
#ifdef SLOWER_MOTION
        float_t stop_x, stop_y;
        float_t dist_x, dist_y;
        float_t reverse_x, reverse_y;
        float_t divisor_x, divisor_y;
        float_t quotient_x, quotient_y;
#endif
#ifdef DEBUG
        float_t prev_x, prev_y;
#endif
        float_t width, height;

        coords.l = START_LEFT;
        coords.r = START_RIGHT;
        coords.b = START_BOTTOM;
        coords.t = START_TOP;
        center_x = coords.l + (coords.r - coords.l)/2.0;
        center_y = coords.b + (coords.t - coords.b)/2.0;
#ifdef DEBUG
        printf("DEBUG: 1st coordinates (%+1.9f,%+1.9f)\n",
                (double)center_x, (double)center_y);
        prev_x = center_x;
        prev_y = center_y;
#endif
#ifdef LINEAR_MOTION
        offset_x = (TARGET_COORDS_X - center_x)/((float_t)NUM_IMAGES);
        offset_y = (TARGET_COORDS_Y - center_y)/((float_t)NUM_IMAGES);
#endif
#ifdef FASTER_MOTION
        start_x = center_x;
        dist_x = TARGET_COORDS_X - start_x;
        if (dist_x >= 0)
            reverse_x = 1.0;
        else
            reverse_x = -1.0;
        factor_x = exp(log(dist_x * reverse_x + 1.0)/((float_t)NUM_IMAGES));
        product_x = 1.0;
        start_y = center_y;
        dist_y = TARGET_COORDS_Y - start_y;
        if (dist_y >= 0)
            reverse_y = 1.0;
        else
            reverse_y = -1.0;
        factor_y = exp(log(dist_y * reverse_y + 1.0)/((float_t)NUM_IMAGES));
        product_y = 1.0;
#endif
#ifdef SLOWER_MOTION
        stop_x = TARGET_COORDS_X;
        dist_x = TARGET_COORDS_X - center_x;
        if (dist_x >= 0)
            reverse_x = 1.0;
        else
            reverse_x = -1.0;
        divisor_x = exp(log(dist_x * reverse_x + 1.0)/((float_t)NUM_IMAGES));
        quotient_x = dist_x * reverse_x + 1.0;
        stop_y = TARGET_COORDS_Y;
        dist_y = TARGET_COORDS_Y - center_y;
        if (dist_y >= 0)
            reverse_y = 1.0;
        else
            reverse_y = -1.0;
        divisor_y = exp(log(dist_y * reverse_y + 1.0)/((float_t)NUM_IMAGES));
        quotient_y = dist_y * reverse_y + 1.0;
#endif

#ifdef DISPLAY
        sprintf(title, "Mandelbrot %d", PAR);
        _SDL_SysC_InitDisplay(WIDTH, HEIGHT, title);
#endif /* DISPLAY */

        for(i=0; i<NUM_IMAGES; i++)
        {
            //t = now() / TIME_BASE;
#ifdef DISPLAY
            if (_SDL_SysC_DisplayClosed())
            {
                printf("Display closed, stimulus quits.\n");
                exit(0);
            }
#endif /* DISPLAY */
#ifdef HELP
            printf("Stimulus sends coordinates set %d.\n", i);
#endif
            //StartTime[i] = t;
            CoordsOut->write(coords);

#ifdef LINEAR_MOTION
            center_x += offset_x;
            center_y += offset_y;
#endif
#ifdef FASTER_MOTION
            product_x *= factor_x;
            product_y *= factor_y;
            center_x = start_x + (product_x-1.0) * reverse_x;
            center_y = start_y + (product_y-1.0) * reverse_y;
#endif
#ifdef SLOWER_MOTION
            quotient_x /= divisor_x;
            quotient_y /= divisor_y;
            center_x = stop_x - (quotient_x-1.0) * reverse_x;
            center_y = stop_y - (quotient_y-1.0) * reverse_y;
#endif
#ifdef DEBUG
            printf("DEBUG: Add coordinates (%+1.9f,%+1.9f)\n",
                    (double)(center_x-prev_x), (double)(center_y-prev_y));
            printf("DEBUG: New coordinates (%+1.9f,%+1.9f)\n",
                    (double)center_x, (double)center_y);
            prev_x = center_x;
            prev_y = center_y;
#endif
            width = (coords.r - coords.l) * (TARGET_ZOOM/2.0);
            height = (coords.t - coords.b) * (TARGET_ZOOM/2.0);
            coords.l = center_x - width;
            coords.r = center_x + width;
            coords.b = center_y - height;
            coords.t = center_y + height;
        }
    }

    SC_CTOR(Stimulus)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(DUT)
{
    sc_port<i_coords_receiver> CoordsIn;
    sc_port<i_image_sender> ImgOut;

    COORDS coords;
    IMAGE image;

    sc_event mandelbrot_run;
    sc_event_and_list main_continue;

#define M(n) sc_event ready_mandel##n
    M(1);
#if PAR > 1
    M(2);
#endif
#if PAR > 2
    M(3); M(4);
#endif
#if PAR > 4
    M(5); M(6); M(7); M(8);
#endif
#if PAR > 8
    M(9); M(10); M(11); M(12); M(13); M(14); M(15); M(16);
#endif
#if PAR > 16
    M(17); M(18); M(19); M(20); M(21); M(22); M(23); M(24);
    M(25); M(26); M(27); M(28); M(29); M(30); M(31); M(32);
#endif
#if PAR > 32
    M(33); M(34); M(35); M(36); M(37); M(38); M(39); M(40);
    M(41); M(42); M(43); M(44); M(45); M(46); M(47); M(48);
    M(49); M(50); M(51); M(52); M(53); M(54); M(55); M(56);
    M(57); M(58); M(59); M(60); M(61); M(62); M(63); M(64);
#endif
#undef M

    static const unsigned char palette[16][3];

// The following four versions of mandlebrot##n are almost the same, differing in combination of two conditions
// Differences between USE_SC_THREAD or not: an infinite top-level loop and wait() vs. next_trigger()
// Differences between DISPLAY or not: invoke SDL APIs vs. not

#ifdef USE_SC_THREAD

#ifdef DISPLAY
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    while(true)                                                                                 \
    {                                                                                           \
        /* The following is taken (with very few adaptations) from: */                          \
        /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                        \
        /* RD, 10/27/12. */                                                                     \
                                                                                                \
        wait(mandelbrot_run);                                                                   \
        /* For each pixel on the screen do: */                                                  \
        for(row = rowStart; row < rowEnd; row++)                                                \
        {                                                                                       \
            for(col = 0; col < WIDTH; col++)                                                    \
            {                                                                                   \
                /* x0 = scaled x coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */ \
                x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);              \
                /* y0 = scaled y coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */   \
                y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);             \
                                                                                                \
                x = 0;                                                                          \
                y = 0;                                                                          \
                                                                                                \
                iteration = 0;                                                                  \
                /* max_iteration = 1000; */                                                     \
                                                                                                \
                while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                        \
                {                                                                               \
                    xtemp = x*x - y*y + x0;                                                     \
                    y = 2*x*y + y0;                                                             \
                                                                                                \
                    x = xtemp;                                                                  \
                                                                                                \
                    iteration = iteration + 1;                                                  \
                }                                                                               \
                                                                                                \
                /* color = iteration */                                                         \
                color = iteration % 16;                                                         \
                                                                                                \
                /* plot(x0,y0,color) */                                                         \
                Image.R[HEIGHT-1-row][col] = palette[color][0];                                 \
                Image.G[HEIGHT-1-row][col] = palette[color][1];                                 \
                Image.B[HEIGHT-1-row][col] = palette[color][2];                                 \
                                                                                                \
                _SDL_SysC_DrawPixel(palette[color][0], palette[color][1],                       \
                                    palette[color][2], col, HEIGHT-1-row);                      \
            }                                                                                   \
            _SDL_SysC_UpdateRect(0, HEIGHT-1-row, WIDTH, 1);                                    \
        }                                                                                       \
                                                                                                \
        ready_mandel##n.notify(SC_ZERO_TIME);                                                   \
    }                                                                                           \
}
#else
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    while(true)                                                                                 \
    {                                                                                           \
        /* The following is taken (with very few adaptations) from: */                          \
        /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                        \
        /* RD, 10/27/12. */                                                                     \
                                                                                                \
        wait(mandelbrot_run);                                                                   \
        /* For each pixel on the screen do: */                                                  \
        for(row = rowStart; row < rowEnd; row++)                                                \
        {                                                                                       \
            for(col = 0; col < WIDTH; col++)                                                    \
            {                                                                                   \
                /* x0 = scaled x coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */ \
                x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);              \
                /* y0 = scaled y coordinate of pixel */                                         \
                /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */   \
                y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);             \
                                                                                                \
                x = 0;                                                                          \
                y = 0;                                                                          \
                                                                                                \
                iteration = 0;                                                                  \
                /* max_iteration = 1000; */                                                     \
                                                                                                \
                while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                        \
                {                                                                               \
                    xtemp = x*x - y*y + x0;                                                     \
                    y = 2*x*y + y0;                                                             \
                                                                                                \
                    x = xtemp;                                                                  \
                                                                                                \
                    iteration = iteration + 1;                                                  \
                }                                                                               \
                                                                                                \
                /* color = iteration */                                                         \
                color = iteration % 16;                                                         \
                                                                                                \
                /* plot(x0,y0,color) */                                                         \
                Image.R[HEIGHT-1-row][col] = palette[color][0];                                 \
                Image.G[HEIGHT-1-row][col] = palette[color][1];                                 \
                Image.B[HEIGHT-1-row][col] = palette[color][2];                                 \
            }                                                                                   \
        }                                                                                       \
                                                                                                \
        ready_mandel##n.notify(SC_ZERO_TIME);                                                   \
    }                                                                                           \
}
#endif /* DISPLAY */

#else /* Use SC_METHOD */

#ifdef DISPLAY
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    /* The following is taken (with very few adaptations) from: */                              \
    /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                            \
    /* RD, 10/27/12. */                                                                         \
                                                                                                \
    /* For each pixel on the screen do: */                                                      \
    for(row = rowStart; row < rowEnd; row++)                                                    \
    {                                                                                           \
        for(col = 0; col < WIDTH; col++)                                                        \
        {                                                                                       \
            /* x0 = scaled x coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */     \
            x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);                  \
            /* y0 = scaled y coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */       \
            y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);                 \
                                                                                                \
            x = 0;                                                                              \
            y = 0;                                                                              \
                                                                                                \
            iteration = 0;                                                                      \
            /* max_iteration = 1000; */                                                         \
                                                                                                \
            while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                            \
            {                                                                                   \
                xtemp = x*x - y*y + x0;                                                         \
                y = 2*x*y + y0;                                                                 \
                                                                                                \
                x = xtemp;                                                                      \
                                                                                                \
                iteration = iteration + 1;                                                      \
            }                                                                                   \
                                                                                                \
            /* color = iteration */                                                             \
            color = iteration % 16;                                                             \
                                                                                                \
            /* plot(x0,y0,color) */                                                             \
            Image.R[HEIGHT-1-row][col] = palette[color][0];                                     \
            Image.G[HEIGHT-1-row][col] = palette[color][1];                                     \
            Image.B[HEIGHT-1-row][col] = palette[color][2];                                     \
                                                                                                \
            _SDL_SysC_DrawPixel(palette[color][0], palette[color][1],                           \
                                palette[color][2], col, HEIGHT-1-row);                          \
        }                                                                                       \
        _SDL_SysC_UpdateRect(0, HEIGHT-1-row, WIDTH, 1);                                        \
    }                                                                                           \
                                                                                                \
    next_trigger();                                                                             \
    ready_mandel##n.notify(SC_ZERO_TIME);                                                       \
}
#else
#define MANDELBROT(n, Coords, rowStart, rowEnd, Image)                                          \
void mandelbrot##n (void)                                                                       \
{                                                                                               \
    float_t x, y;                                                                               \
    float_t x0, y0;                                                                             \
    float_t xtemp;                                                                              \
    unsigned int iteration;                                                                     \
    const unsigned int max_iteration = MAX_ITER;                                                \
    unsigned char color;                                                                        \
    int row, col;                                                                               \
                                                                                                \
    /* The following is taken (with very few adaptations) from: */                              \
    /* http://en.wikipedia.org/wiki/Mandelbrot_set#For_programmers*/                            \
    /* RD, 10/27/12. */                                                                         \
                                                                                                \
    /* For each pixel on the screen do: */                                                      \
    for(row = rowStart; row < rowEnd; row++)                                                    \
    {                                                                                           \
        for(col = 0; col < WIDTH; col++)                                                        \
        {                                                                                       \
            /* x0 = scaled x coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot X scale (-2.5, 1) */     \
            x0 = Coords.l + (Coords.r-Coords.l)*((float_t)col/(float_t)WIDTH);                  \
            /* y0 = scaled y coordinate of pixel */                                             \
            /*      (must be scaled to lie somewhere in the mandelbrot Y scale (-1, 1) */       \
            y0 = Coords.b + (Coords.t-Coords.b)*((float_t)row/(float_t)HEIGHT);                 \
                                                                                                \
            x = 0;                                                                              \
            y = 0;                                                                              \
                                                                                                \
            iteration = 0;                                                                      \
            /* max_iteration = 1000; */                                                         \
                                                                                                \
            while ( x*x + y*y < 2*2  &&  iteration < max_iteration )                            \
            {                                                                                   \
                xtemp = x*x - y*y + x0;                                                         \
                y = 2*x*y + y0;                                                                 \
                                                                                                \
                x = xtemp;                                                                      \
                                                                                                \
                iteration = iteration + 1;                                                      \
            }                                                                                   \
                                                                                                \
            /* color = iteration */                                                             \
            color = iteration % 16;                                                             \
                                                                                                \
            /* plot(x0,y0,color) */                                                             \
            Image.R[HEIGHT-1-row][col] = palette[color][0];                                     \
            Image.G[HEIGHT-1-row][col] = palette[color][1];                                     \
            Image.B[HEIGHT-1-row][col] = palette[color][2];                                     \
        }                                                                                       \
    }                                                                                           \
                                                                                                \
    next_trigger();                                                                             \
    ready_mandel##n.notify(SC_ZERO_TIME);                                                       \
}
#endif /* DISPLAY */

#endif /* USE_SC_THREAD */

#define M(n) MANDELBROT(n, coords, (HEIGHT/PAR*(n-1)), (HEIGHT/PAR*n), image)
    M(1) 
#if PAR > 1
    M(2) 
#endif
#if PAR > 2
    M(3) M(4) 
#endif
#if PAR > 4
    M(5) M(6) M(7) M(8) 
#endif
#if PAR > 8
    M(9) M(10) M(11) M(12) M(13) M(14) M(15) M(16)
#endif
#if PAR > 16
    M(17) M(18) M(19) M(20) M(21) M(22) M(23) M(24)
    M(25) M(26) M(27) M(28) M(29) M(30) M(31) M(32)
#endif
#if PAR > 32
    M(33) M(34) M(35) M(36) M(37) M(38) M(39) M(40)
    M(41) M(42) M(43) M(44) M(45) M(46) M(47) M(48)
    M(49) M(50) M(51) M(52) M(53) M(54) M(55) M(56)
    M(57) M(58) M(59) M(60) M(61) M(62) M(63) M(64)
#endif
#undef M

#undef MANDELBROT

    void main(void)
    {
        while(1)
        {
            CoordsIn->read(coords);
#ifdef DISPLAY
            // _SDL_SysC_ClearDisplay();
            _SDL_SysC_FillDisplay(42,42,42);
#endif /* DISPLAY */
            mandelbrot_run.notify(SC_ZERO_TIME);
            wait(ready_mandel1);
            ImgOut->write(image);
        }
    }

    void before_end_of_elaboration()
    {
        // initialize sc_event_and_list main_continue
        #define M(n) ready_mandel##n
                main_continue &= M(1);
        #if PAR >1
                main_continue &= M(2);
        #endif
        #if PAR > 2
                main_continue &= M(3);
                main_continue &= M(4);
        #endif
        #if PAR > 4
                main_continue &= M(5);
                main_continue &= M(6);
                main_continue &= M(7);
                main_continue &= M(8);
        #endif
        #if PAR > 8
                main_continue &= M(9);
                main_continue &= M(10);
                main_continue &= M(11);
                main_continue &= M(12);
                main_continue &= M(13);
                main_continue &= M(14);
                main_continue &= M(15);
                main_continue &= M(16);
        #endif
        #if PAR > 16
                main_continue &= M(17);
                main_continue &= M(18);
                main_continue &= M(19);
                main_continue &= M(20);
                main_continue &= M(21);
                main_continue &= M(22);
                main_continue &= M(23);
                main_continue &= M(24);
                main_continue &= M(25);
                main_continue &= M(26);
                main_continue &= M(27);
                main_continue &= M(28);
                main_continue &= M(29);
                main_continue &= M(30);
                main_continue &= M(31);
                main_continue &= M(32);
        #endif
        #if PAR > 32
                main_continue &= M(33);
                main_continue &= M(34);
                main_continue &= M(35);
                main_continue &= M(36);
                main_continue &= M(37);
                main_continue &= M(38);
                main_continue &= M(39);
                main_continue &= M(40);
                main_continue &= M(41);
                main_continue &= M(42);
                main_continue &= M(43);
                main_continue &= M(44);
                main_continue &= M(45);
                main_continue &= M(46);
                main_continue &= M(47);
                main_continue &= M(48);
                main_continue &= M(49);
                main_continue &= M(50);
                main_continue &= M(51);
                main_continue &= M(52);
                main_continue &= M(53);
                main_continue &= M(54);
                main_continue &= M(55);
                main_continue &= M(56);
                main_continue &= M(57);
                main_continue &= M(58);
                main_continue &= M(59);
                main_continue &= M(60);
                main_continue &= M(61);
                main_continue &= M(62);
                main_continue &= M(63);
                main_continue &= M(64);
        #endif
        #undef M
        // end of initialization of main_continue
    }

    SC_CTOR(DUT)
    {
        // construct main thread
        SC_THREAD(main);

        // construct worker methods
#ifdef USE_SC_THREAD
#define M(n) SC_THREAD(mandelbrot##n); sensitive << mandelbrot_run;
#else
#define M(n) SC_METHOD(mandelbrot##n); sensitive << mandelbrot_run;
#endif /* USE_SC_THREAD */
        M(1);
#if PAR > 1
        M(2);
#endif
#if PAR > 2
        M(3); M(4);
#endif
#if PAR > 4
        M(5); M(6); M(7); M(8);
#endif
#if PAR > 8
        M(9); M(10); M(11); M(12); M(13); M(14); M(15); M(16);
#endif
#if PAR > 16
        M(17); M(18); M(19); M(20); M(21); M(22); M(23); M(24);
        M(25); M(26); M(27); M(28); M(29); M(30); M(31); M(32);
#endif
#if PAR > 32
        M(33); M(34); M(35); M(36); M(37); M(38); M(39); M(40);
        M(41); M(42); M(43); M(44); M(45); M(46); M(47); M(48);
        M(49); M(50); M(51); M(52); M(53); M(54); M(55); M(56);
        M(57); M(58); M(59); M(60); M(61); M(62); M(63); M(64);
#endif
#undef M
    }
};

const unsigned char DUT::palette[16][3] = {
    //  r    g    b
    {   0,   0,   0 },  // 0, black
    { 127,   0,   0 },  // 1, brown
    { 255,   0,   0 },  // 2, red
    { 255, 127,   0 },  // 3, orange
    { 255, 255,   0 },  // 4, yellow
    { 127, 255,   0 },  // 5, light green
    {   0, 255,   0 },  // 6, green
    {   0, 255, 127 },  // 7, blue green
    {   0, 255, 255 },  // 8, turquoise
    { 127, 255, 255 },  // 9, light blue
    { 255, 255, 255 },  // 10, white
    { 255, 127, 255 },  // 11, pink
    { 255,   0, 255 },  // 12, light pink
    { 127,   0, 255 },  // 13, purple
    {   0,   0, 255 },  // 14, blue
    {   0,   0, 127 }}; // 15, dark blue


SC_MODULE(DataIn)
{
    sc_port<i_coords_receiver> CoordsIn;
    sc_port<i_coords_sender> CoordsOut;

    COORDS coords;

    void main(void)
    {
        while(1)
        {
            CoordsIn->read(coords);
            CoordsOut->write(coords);
        }
    }

    SC_CTOR(DataIn)
    {
        SC_THREAD(main);
    }
};


SC_MODULE(DataOut)
{
    sc_port<i_image_receiver> ImgIn;
    sc_port<i_image_sender> ImgOut;

    IMAGE image;

    void main(void)
    {
        while(1)
        {
            ImgIn->read(image);
            ImgOut->write(image);
        }
    }

    SC_CTOR(DataOut)
    {
        SC_THREAD(main);
#ifdef SYSC_TYPE_CHNL
        set_stack_size(128*1024*1024);
#endif /* SYSC_TYPE_CHNL */
    }
};


SC_MODULE(Platform)
{
    sc_port<i_coords_receiver> CoordsIn;
    sc_port<i_image_sender> ImgOut;

    c_coords_channel   platform_c1;
    c_image_channel    platform_c2;
    DataIn          din;
    DUT             dut;
    DataOut         dout;

    SC_CTOR(Platform):
        platform_c1("platform_c1"),
        platform_c2("platform_c2"),
        din("din"),
        dut("dut"),
        dout("dout")
    {
        // port binding
        //din(CoordsIn, platform_c1);
        //dut(platform_c1, platform_c2);
        //dout(platform_c2, ImgOut);
    }

    void before_end_of_elaboration()
    {
        din.CoordsIn.bind(CoordsIn);
        din.CoordsOut.bind(platform_c1);

        dut.CoordsIn.bind(platform_c1);
        dut.ImgOut.bind(platform_c2);

        dout.ImgIn.bind(platform_c2);
        dout.ImgOut.bind(ImgOut);
    }
};


SC_MODULE(TOP)
{
    c_coords_channel   c1;
    c_image_channel    c2;

    Stimulus    stimulus;
    Platform    platform;
    Monitor     monitor;

    SC_CTOR(TOP):
        c1("c1"),
        c2("c2"),
        stimulus("stimulus"),
        platform("platform"),
        monitor("monitor")
    {
        // port binding
        //stimulus(c1);
        //platform(c1, c2);
        //monitor(c2);
    }

    void before_end_of_elaboration()
    {
        stimulus.CoordsOut.bind(c1);

        platform.CoordsIn.bind(c1);
        platform.ImgOut.bind(c2);

        monitor.ImgIn.bind(c2);
    }
};


int sc_main(int argc, char* argv[])
{
    TOP top("top");
#ifdef HELP
    printf("Mandelbrot starts.\n");
#endif
    sc_start();
    assert(false); // never reach here
#ifdef HELP
    printf("Mandlebrot done.\n");
#endif
    return 0;
}

// EOF
