#ifndef TIME_ADVANCED_TABLE_H_INCLUDED_
#define TIME_ADVANCED_TABLE_H_INCLUDED_

#include "segment_graph.h"
#include "segment.h"
#include "time.h"

namespace risc {

namespace sg {

/**
 * \class TimeAdvancedTable
 * \brief This class represents the time advanced tables for a given segment
 *        graph. We distinguish between the 'current time table' and the
 *        'next time table'.
 */
class TimeAdvancedTable {

private:

  typedef boost::graph_traits<risc::sg::Graph>::vertex_iterator VertexIterator;

public:
  /**
   * \brief    Default constructor for the TimeAdvancedTable
   */
  TimeAdvancedTable(SegmentGraph &graph);

  /**
   * \brief    Default destructor for the TimeAdvancedTable
   */
  ~TimeAdvancedTable();

  /**
   * \function determine_conflict_table
   * \brief    This function determines conflicts among the segments
   */
  void determine_conflict_table();

  /**
   * \function print_table
   * \brief    This function prints the current time advance table in a html
   *           file
   */
  void print_current_time_advanced_table(std::string filename);

  /**
   * \function print_table
   * \brief    This function prints the next time advance table in a html file
   */
  void print_next_time_advanced_table(std::string filename);

  /**
   * \function print_time_table
   * \brief    This function is only for internal use. It is used as the
   *           base function for the other print functions
   */
  void print_time_table(Time *table, std::string filename);

  /**
   * \function get_waiting_time_for_segment
   * \brief    This function determines how much time is spend while entering a
   *           segment
   */
  Time get_waiting_time_for_segment(VertexIterator Segment);

  /**
   * \brief    Reference to the associated segment graph of the conflict table
   */
  SegmentGraph &segment_graph_;

  /**
   * \brief    Size of the conflict table
   */
  int size_;

  /**
   * \brief    This vector is a lookup between the conflict table index and the
   *           associated segment index
   */
  std::vector<VertexIterator> table_index_segment_lookup_;

  /**
   * \function get_segment_id
   * \brief    This function looks up which segment id belongs to a given table
   *           index
   */
  int get_segment_id(int index);

  /**
   * \function convert_si_magnitude
   * \brief    This function converts a given string like "SC_SEC" into the
   *           corresponding amount of femto seconds
   */
  long long convert_si_magnitude(std::string magnitude);

  /**
   * \brief    Pointer for the current and next time advance table
   */
  Time *current_time_advance_;
  Time *next_time_advance_;
};

}; // end of namesapce sg

}; // end of namespace risc

#endif /* TIME_ADVANCED_TABLE_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
