#ifndef MISC_H_INCLUDED_
#define MISC_H_INCLUDED_

#include "rose.h"
#include <string>

namespace risc
{

namespace tools
{

/**
 * \brief               This function checks if the passed type
 *                      represents a data structure
 * \param type          Type that will be checked
 * \return              True, if the type represents a data structure
 */
bool
is_data_structure(SgType* type);

/**
 * \brief               This function checks if the passed type
 *                      represents a C++ builtin type
 * \param type          Type that will be checked
 * \return              True, if the type represents a C++ builtin type
 */
bool
is_builtin_type(SgType* type);


/**
 * \brief               Compares two classes
 * \param cd1           Class definition 1
 * \param cd2           Class definition 2
 * \return              True, if the types are equal
 */
bool
class_types_are_identical(SgClassDefinition* cd1, SgClassDefinition* cd2);

/**
 * \brief               Compares the function name with the given string
 * \param function      The function declaration
 * \param name          String which will be compared
 * \return              True, if the string and the function name are equal
 */
bool
is_func_name(const SgFunctionDeclaration* function, const std::string& name);

/**
 * \brief               Prints a string with indention
 * \param s             Sting to print
 * \param indent        Indention level
 */
void
print_indented(std::string s, unsigned indent);

/**
 * \brief               Determines the class defintion of varialbe
 *                      It is expected that the type is SgClassType
 * \param var_decl      The given variable declaration
 * \return              Associated SgClassDefintion of the var_decl
 */
SgClassDefinition*
class_definition_of_variable(SgVariableDeclaration* var_decl);

/**
 * \brief               Checks if a given function is a C++ builtin function
 *                      This function is taken/inspired from the Rose
 *                      builtinFilter structure
 * \param func_decl     Function declaration
 * \return              true, if builtin function
 */
bool
is_builtin_function(SgFunctionDeclaration *func_decl);

/**
 * \function all_cfg_branches_end_in_break
 * \brief    This function determines if all cfg branches in the statement
 *           'node' will end in a break statement.
 *           The statement in end will be excluded. Some examples:
 *
 *           input with result: true
 *           break;  <-- node
 *           int a;
 *
 *           input with result: false
 *           if(test)  <-- node
 *             break;
 *           int a;
 *
 *           input with result: true
 *           if(test)  <-- node
 *             break;
 *           else
 *             break;
 *           int a;
 */
bool
all_cfg_branches_end_in_break(SgNode *node, int &break_counter);

/**
 * \function all_cfg_branches_end_in_break
 * \brief    This function determines if all cfg branches in the statement
 *           'node' will end in a continue statement.
 *           The statement in end will be excluded. Some examples:
 *
 *           input with result: true
 *           continue;  <-- node
 *           int a;
 *
 *           input with result: false
 *           if(test);  <-- node
 *             continue;
 *           int a;
 *
 *           input with result: true
 *           if(test);  <-- node
 *             continue;
 *           else
 *             continue;
 *           int a;
 */
bool
all_cfg_branches_end_in_continue(SgNode *node, int &continue_counter);

/**
 * \function get_previous_stmt_in_basic_block
 * \brief    This function determines the previous statement of
 *           the statement 'node' in the next surrounding basic block.
 *           If 'node' is the first statement, the return value will be
 *           NULL.
 */
SgStatement*
get_previous_stmt_in_basic_block(SgNode *node);

/**
 * \function is_conditional_break
 * \brief    This function checks if a given break statement is unconditioned
 *
 * \param    loop_stmt  Pointer to the loop statement
 * \param    break_stmt Pointer to the break statement
 */
bool
is_conditional_break(SgScopeStatement const * const loop_stmt,
    SgBreakStmt const * const break_stmt);

/**
 * \function find_break_associated_loop
 * \brief    This function finds the surrounding scope statement of a break.
 *           The break can be surrounded by a while, do while, for loop or
 *           by a switch statement.
 */
SgScopeStatement*
find_break_surrounding_scope(SgBreakStmt *break_stmt);

/**
 * \function break_belongs_to_loop
 * \brief    This function checks if a given break statement belongs to a
 *           given loop statements.
 *
 * \param    loop_stmt  Pointer to the loop statement
 * \param    break_stmt Pointer to the break statement
 */
bool
break_belongs_to_loop(SgScopeStatement *loop_stmt,
    SgBreakStmt const * const break_stmt);

bool is_located_in_rose_header(SgNode *node);

/**
 * \function get_constructor
 * \brief    This function returns the constructor of class.
 *           For a module, primitive channel, and a hierarchical channel we
 *           expect exactly one constructor. So, we return the first
 *           SgFunctionDefinition which es a constructor
 */
SgMemberFunctionDeclaration*
get_constructor(SgClassDefinition *class_def);

}; // end namespace tools

}; // end namespace risc

#endif /* MISC_H_INCLUDED_ */

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
